"""

hub.exec.boto3.client.events.delete_rule
hub.exec.boto3.client.events.describe_rule
hub.exec.boto3.client.events.disable_rule
hub.exec.boto3.client.events.enable_rule
hub.exec.boto3.client.events.list_rules
hub.exec.boto3.client.events.put_rule
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    targets: List[
        make_dataclass(
            "Target",
            [
                ("Id", str),
                ("Arn", str),
                ("RoleArn", str, field(default=None)),
                ("Input", str, field(default=None)),
                ("InputPath", str, field(default=None)),
                (
                    "InputTransformer",
                    make_dataclass(
                        "InputTransformer",
                        [
                            ("InputTemplate", str),
                            ("InputPathsMap", Dict[str, str], field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "KinesisParameters",
                    make_dataclass("KinesisParameters", [("PartitionKeyPath", str)]),
                    field(default=None),
                ),
                (
                    "RunCommandParameters",
                    make_dataclass(
                        "RunCommandParameters",
                        [
                            (
                                "RunCommandTargets",
                                List[
                                    make_dataclass(
                                        "RunCommandTarget",
                                        [("Key", str), ("Values", List[str])],
                                    )
                                ],
                            )
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "EcsParameters",
                    make_dataclass(
                        "EcsParameters",
                        [
                            ("TaskDefinitionArn", str),
                            ("TaskCount", int, field(default=None)),
                            ("LaunchType", str, field(default=None)),
                            (
                                "NetworkConfiguration",
                                make_dataclass(
                                    "NetworkConfiguration",
                                    [
                                        (
                                            "awsvpcConfiguration",
                                            make_dataclass(
                                                "AwsVpcConfiguration",
                                                [
                                                    ("Subnets", List[str]),
                                                    (
                                                        "SecurityGroups",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "AssignPublicIp",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                            field(default=None),
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("PlatformVersion", str, field(default=None)),
                            ("Group", str, field(default=None)),
                            (
                                "CapacityProviderStrategy",
                                List[
                                    make_dataclass(
                                        "CapacityProviderStrategyItem",
                                        [
                                            ("capacityProvider", str),
                                            ("weight", int, field(default=None)),
                                            ("base", int, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("EnableECSManagedTags", bool, field(default=None)),
                            ("EnableExecuteCommand", bool, field(default=None)),
                            (
                                "PlacementConstraints",
                                List[
                                    make_dataclass(
                                        "PlacementConstraint",
                                        [
                                            ("type", str, field(default=None)),
                                            ("expression", str, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            (
                                "PlacementStrategy",
                                List[
                                    make_dataclass(
                                        "PlacementStrategy",
                                        [
                                            ("type", str, field(default=None)),
                                            ("field", str, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("PropagateTags", str, field(default=None)),
                            ("ReferenceId", str, field(default=None)),
                            (
                                "Tags",
                                List[
                                    make_dataclass(
                                        "Tag", [("Key", str), ("Value", str)]
                                    )
                                ],
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "BatchParameters",
                    make_dataclass(
                        "BatchParameters",
                        [
                            ("JobDefinition", str),
                            ("JobName", str),
                            (
                                "ArrayProperties",
                                make_dataclass(
                                    "BatchArrayProperties",
                                    [("Size", int, field(default=None))],
                                ),
                                field(default=None),
                            ),
                            (
                                "RetryStrategy",
                                make_dataclass(
                                    "BatchRetryStrategy",
                                    [("Attempts", int, field(default=None))],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "SqsParameters",
                    make_dataclass(
                        "SqsParameters", [("MessageGroupId", str, field(default=None))]
                    ),
                    field(default=None),
                ),
                (
                    "HttpParameters",
                    make_dataclass(
                        "HttpParameters",
                        [
                            ("PathParameterValues", List[str], field(default=None)),
                            ("HeaderParameters", Dict[str, str], field(default=None)),
                            (
                                "QueryStringParameters",
                                Dict[str, str],
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "RedshiftDataParameters",
                    make_dataclass(
                        "RedshiftDataParameters",
                        [
                            ("Sql", str),
                            ("Database", str),
                            ("SecretManagerArn", str, field(default=None)),
                            ("DbUser", str, field(default=None)),
                            ("StatementName", str, field(default=None)),
                            ("WithEvent", bool, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "SageMakerPipelineParameters",
                    make_dataclass(
                        "SageMakerPipelineParameters",
                        [
                            (
                                "PipelineParameterList",
                                List[
                                    make_dataclass(
                                        "SageMakerPipelineParameter",
                                        [("Name", str), ("Value", str)],
                                    )
                                ],
                                field(default=None),
                            )
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "DeadLetterConfig",
                    make_dataclass(
                        "DeadLetterConfig", [("Arn", str, field(default=None))]
                    ),
                    field(default=None),
                ),
                (
                    "RetryPolicy",
                    make_dataclass(
                        "RetryPolicy",
                        [
                            ("MaximumRetryAttempts", int, field(default=None)),
                            ("MaximumEventAgeInSeconds", int, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
            ],
        )
    ] = None,
    schedule_expression: str = None,
    event_pattern: str = None,
    rule_status: str = None,
    description: str = None,
    role_arn: str = None,
    event_bus_name: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    r"""

    Enables the specified rule. If the rule does not exist, the operation fails. When you enable a rule, incoming
    events might not immediately start matching to a newly enabled rule. Allow a short period of time for changes to
    take effect.

    Args:
        hub:
        ctx:
        name(Text): A name, ID to identify the resource.
        resource_id(Text, Optional): The name of the AWS CloudWatch Events Rule.
        targets(List[Dict[str, Any]]): The targets to update or add to the rule.
            * Id (str): The ID of the target within the specified rule. Use this ID to reference the target when
                updating the rule. We recommend using a memorable and unique string.
            * Arn (str): The Amazon Resource Name (ARN) of the target.
            * RoleArn (str, optional): The Amazon Resource Name (ARN) of the IAM role to be used for this target when the rule is
                triggered. If one rule triggers multiple targets, you can use a different IAM role for each
                target.
            * Input (str, optional): Valid JSON text passed to the target. In this case, nothing from the event itself is passed to
                the target. For more information, see The JavaScript Object Notation (JSON) Data Interchange
                Format.
            * InputPath (str, optional): The value of the JSONPath that is used for extracting part of the matched event when passing it
                to the target. You must use JSON dot notation, not bracket notation. For more information about
                JSON paths, see JSONPath.
            * InputTransformer (Dict[str, Any], optional): Settings to enable you to provide custom input to a target based on certain event data. You can
                extract one or more key-value pairs from the event and then use that data to send customized
                input to the target.
                * InputPathsMap (Dict[str, str], optional): Map of JSON paths to be extracted from the event. You can then insert these in the template in
            InputTemplate to produce the output you want to be sent to the target.  InputPathsMap is an
            array key-value pairs, where each value is a valid JSON path. You can have as many as 100 key-
            value pairs. You must use JSON dot notation, not bracket notation. The keys cannot start with
            "Amazon Web Services."
                * InputTemplate (str): Input template where you specify placeholders that will be filled with the values of the keys
                    from InputPathsMap to customize the data sent to the target. Enclose each InputPathsMaps value
                    in brackets: <value> The InputTemplate must be valid JSON. If InputTemplate is a JSON object
                    (surrounded by curly braces), the following restrictions apply:   The placeholder cannot be used
                    as an object key.   The following example shows the syntax for using InputPathsMap and
                    InputTemplate.
                       "InputTransformer":
                       {
                            "InputPathsMap": {"instance": "$.detail.instance","status": "$.detail.status"},
                            "InputTemplate": "<instance> is in state <status>"
                       }
                    To have the InputTemplate include quote marks within a JSON string, escape each
                    quote marks with a slash, as in the following example:
                       "InputTransformer":   {
                            "InputPathsMap": {"instance": "$.detail.instance","status": "$.detail.status"},
                            "InputTemplate": "<instance> is in state \"<status>\""
                            }
                    The InputTemplate can also be valid JSON with varibles in quotes or out, as in the following example:
                       "InputTransformer":   {
                            "InputPathsMap": {"instance": "$.detail.instance","status": "$.detail.status"},
                            "InputTemplate": '{"myInstance": <instance>,"myStatus": "<instance> is in state \"<status>\""}'
                        }
            * KinesisParameters (Dict[str, Any], optional): The custom parameter you can use to control the shard assignment, when the target is a Kinesis
                data stream. If you do not include this parameter, the default is to use the eventId as the
                partition key.
                * PartitionKeyPath (str): The JSON path to be extracted from the event and used as the partition key. For more
                    information, see Amazon Kinesis Streams Key Concepts in the Amazon Kinesis Streams Developer
                    Guide.
            * RunCommandParameters (Dict[str, Any], optional): Parameters used when you are using the rule to invoke Amazon EC2 Run Command.
                * RunCommandTargets (List[Dict[str, Any]]): Currently, we support including only one RunCommandTarget block, which specifies either an array
                    of InstanceIds or a tag.
                    * Key (str): Can be either tag: tag-key or InstanceIds.
                    * Values (List[str]): If Key is tag: tag-key, Values is a list of tag values. If Key is InstanceIds, Values is a list
                        of Amazon EC2 instance IDs.
            * EcsParameters (Dict[str, Any], optional): Contains the Amazon ECS task definition and task count to be used, if the event target is an
                Amazon ECS task. For more information about Amazon ECS tasks, see Task Definitions  in the
                Amazon EC2 Container Service Developer Guide.
                * TaskDefinitionArn (str): The ARN of the task definition to use if the event target is an Amazon ECS task.
                * TaskCount (int, optional): The number of tasks to create based on TaskDefinition. The default is 1.
                * LaunchType (str, optional): Specifies the launch type on which your task is running. The launch type that you specify here
                    must match one of the launch type (compatibilities) of the target task. The FARGATE value is
                    supported only in the Regions where Fargate with Amazon ECS is supported. For more information,
                    see Fargate on Amazon ECS in the Amazon Elastic Container Service Developer Guide.
                * NetworkConfiguration (Dict[str, Any], optional): Use this structure if the Amazon ECS task uses the awsvpc network mode. This structure specifies
                    the VPC subnets and security groups associated with the task, and whether a public IP address is
                    to be used. This structure is required if LaunchType is FARGATE because the awsvpc mode is
                    required for Fargate tasks. If you specify NetworkConfiguration when the target ECS task does
                    not use the awsvpc network mode, the task fails.
                    * awsvpcConfiguration (Dict[str, Any], optional): Use this structure to specify the VPC subnets and security groups for the task, and whether a
                        public IP address is to be used. This structure is relevant only for ECS tasks that use the
                        awsvpc network mode.
                        * Subnets (List[str]): Specifies the subnets associated with the task. These subnets must all be in the same VPC. You
                            can specify as many as 16 subnets.
                        * SecurityGroups (List[str], optional): Specifies the security groups associated with the task. These security groups must all be in the
                            same VPC. You can specify as many as five security groups. If you do not specify a security
                            group, the default security group for the VPC is used.
                        * AssignPublicIp (str, optional): Specifies whether the task's elastic network interface receives a public IP address. You can
                            specify ENABLED only when LaunchType in EcsParameters is set to FARGATE.
                * PlatformVersion (str, optional): Specifies the platform version for the task. Specify only the numeric portion of the platform
                    version, such as 1.1.0. This structure is used only if LaunchType is FARGATE. For more
                    information about valid platform versions, see Fargate Platform Versions in the Amazon Elastic
                    Container Service Developer Guide.
                * Group (str, optional): Specifies an ECS task group for the task. The maximum length is 255 characters.
                * CapacityProviderStrategy (List[Dict[str, Any]], optional): The capacity provider strategy to use for the task. If a capacityProviderStrategy is specified,
                    the launchType parameter must be omitted. If no capacityProviderStrategy or launchType is
                    specified, the defaultCapacityProviderStrategy for the cluster is used.
                    * capacityProvider (str): The short name of the capacity provider.
                    * weight (int, optional): The weight value designates the relative percentage of the total number of tasks launched that
                        should use the specified capacity provider. The weight value is taken into consideration after
                        the base value, if defined, is satisfied.
                    * base (int, optional): The base value designates how many tasks, at a minimum, to run on the specified capacity
                        provider. Only one capacity provider in a capacity provider strategy can have a base defined. If
                        no value is specified, the default value of 0 is used.
                * EnableECSManagedTags (bool, optional): Specifies whether to enable Amazon ECS managed tags for the task. For more information, see
                    Tagging Your Amazon ECS Resources in the Amazon Elastic Container Service Developer Guide.
                * EnableExecuteCommand (bool, optional): Whether or not to enable the execute command functionality for the containers in this task. If
                    true, this enables execute command functionality on all containers in the task.
                * PlacementConstraints (List[Dict[str, Any]], optional): An array of placement constraint objects to use for the task. You can specify up to 10
                    constraints per task (including constraints in the task definition and those specified at
                    runtime).
                    * type (str, optional): The type of constraint. Use distinctInstance to ensure that each task in a particular group is
                        running on a different container instance. Use memberOf to restrict the selection to a group of
                        valid candidates.
                    * expression (str, optional): A cluster query language expression to apply to the constraint. You cannot specify an expression
                        if the constraint type is distinctInstance. To learn more, see Cluster Query Language in the
                        Amazon Elastic Container Service Developer Guide.
                * PlacementStrategy (List[Dict[str, Any]], optional): The placement strategy objects to use for the task. You can specify a maximum of five strategy
                    rules per task.
                    * type (str, optional): The type of placement strategy. The random placement strategy randomly places tasks on available
                        candidates. The spread placement strategy spreads placement across available candidates evenly
                        based on the field parameter. The binpack strategy places tasks on available candidates that
                        have the least available amount of the resource that is specified with the field parameter. For
                        example, if you binpack on memory, a task is placed on the instance with the least amount of
                        remaining memory (but still enough to run the task).
                    * field (str, optional): The field to apply the placement strategy against. For the spread placement strategy, valid
                        values are instanceId (or host, which has the same effect), or any platform or custom attribute
                        that is applied to a container instance, such as attribute:ecs.availability-zone. For the
                        binpack placement strategy, valid values are cpu and memory. For the random placement strategy,
                        this field is not used.
                * PropagateTags (str, optional): Specifies whether to propagate the tags from the task definition to the task. If no value is
                    specified, the tags are not propagated. Tags can only be propagated to the task during task
                    creation. To add tags to a task after task creation, use the TagResource API action.
                * ReferenceId (str, optional): The reference ID to use for the task.
                * Tags (List[Dict[str, Any]], optional): The metadata that you apply to the task to help you categorize and organize them. Each tag
                    consists of a key and an optional value, both of which you define. To learn more, see RunTask in
                    the Amazon ECS API Reference.
                    * Key (str): A string you can use to assign a value. The combination of tag keys and values can help you
            organize and categorize your resources.
                    * Value (str): The value for the specified tag key.
            * BatchParameters (Dict[str, Any], optional): If the event target is an Batch job, this contains the job definition, job name, and other
                parameters. For more information, see Jobs in the Batch User Guide.
                * JobDefinition (str): The ARN or name of the job definition to use if the event target is an Batch job. This job
                    definition must already exist.
                * JobName (str): The name to use for this execution of the job, if the target is an Batch job.
                * ArrayProperties (Dict[str, Any], optional): The array properties for the submitted job, such as the size of the array. The array size can be
                    between 2 and 10,000. If you specify array properties for a job, it becomes an array job. This
                    parameter is used only if the target is an Batch job.
                    * Size (int, optional): The size of the array, if this is an array batch job. Valid values are integers between 2 and
                        10,000.
                * RetryStrategy (Dict[str, Any], optional): The retry strategy to use for failed jobs, if the target is an Batch job. The retry strategy is
                    the number of times to retry the failed job execution. Valid values are 1–10. When you specify a
                    retry strategy here, it overrides the retry strategy defined in the job definition.
                    * Attempts (int, optional): The number of times to attempt to retry, if the job fails. Valid values are 1–10.
            * SqsParameters (Dict[str, Any], optional): Contains the message group ID to use when the target is a FIFO queue. If you specify an SQS FIFO
                queue as a target, the queue must have content-based deduplication enabled.
                * MessageGroupId (str, optional): The FIFO message group ID to use as the target.
            * HttpParameters (Dict[str, Any], optional): Contains the HTTP parameters to use when the target is a API Gateway REST endpoint or
                EventBridge ApiDestination. If you specify an API Gateway REST API or EventBridge ApiDestination
                as a target, you can use this parameter to specify headers, path parameters, and query string
                keys/values as part of your target invoking request. If you're using ApiDestinations, the
                corresponding Connection can also have these values configured. In case of any conflicting keys,
                values from the Connection take precedence.
                * PathParameterValues (List[str], optional): The path parameter values to be used to populate API Gateway REST API or EventBridge
                    ApiDestination path wildcards ("*").
                * HeaderParameters (Dict[str, str], optional): The headers that need to be sent as part of request invoking the API Gateway REST API or
                    EventBridge ApiDestination.
                * QueryStringParameters (Dict[str, str], optional): The query string keys/values that need to be sent as part of request invoking the API Gateway
                    REST API or EventBridge ApiDestination.
            * RedshiftDataParameters (Dict[str, Any], optional): Contains the Amazon Redshift Data API parameters to use when the target is a Amazon Redshift
                cluster. If you specify a Amazon Redshift Cluster as a Target, you can use this to specify
                parameters to invoke the Amazon Redshift Data API ExecuteStatement based on EventBridge events.
                * SecretManagerArn (str, optional): The name or ARN of the secret that enables access to the database. Required when authenticating
                    using Amazon Web Services Secrets Manager.
                * Database (str): The name of the database. Required when authenticating using temporary credentials.
                * DbUser (str, optional): The database user name. Required when authenticating using temporary credentials.
                * Sql (str): The SQL statement text to run.
                * StatementName (str, optional): The name of the SQL statement. You can name the SQL statement when you create it to identify the
                    query.
                * WithEvent (bool, optional): Indicates whether to send an event back to EventBridge after the SQL statement runs.
            * SageMakerPipelineParameters (Dict[str, Any], optional): Contains the SageMaker Model Building Pipeline parameters to start execution of a SageMaker
                Model Building Pipeline. If you specify a SageMaker Model Building Pipeline as a target, you can
                use this to specify parameters to start a pipeline execution based on EventBridge events.
                * PipelineParameterList (List[Dict[str, Any]], optional): List of Parameter names and values for SageMaker Model Building Pipeline execution.
                    * Name (str): Name of parameter to start execution of a SageMaker Model Building Pipeline.
                    * Value (str): Value of parameter to start execution of a SageMaker Model Building Pipeline.
            * DeadLetterConfig (Dict[str, Any], optional): The DeadLetterConfig that defines the target queue to send dead-letter queue events to.
                * Arn (str, optional): The ARN of the SQS queue specified as the target for the dead-letter queue.
            * RetryPolicy (Dict[str, Any], optional): The RetryPolicy object that contains the retry policy configuration to use for the dead-letter
                queue.
                * MaximumRetryAttempts (int, optional): The maximum number of retry attempts to make before the request fails. Retry attempts continue
                    until either the maximum number of attempts is made or until the duration of the
                    MaximumEventAgeInSeconds is met.
                * MaximumEventAgeInSeconds (int, optional): The maximum amount of time, in seconds, to continue to make retry attempts.
        schedule_expression (Text, Optional): Scheduling expression.
            For example, "cron(0 20 * * ? *)" or "rate(5 minutes)".
        event_pattern (Text, Optional): Rules use event patterns to select events and route them to targets. A pattern
            either matches an event or it doesn't. Event patterns are represented as JSON objects with a structure that
            is similar to that of events.
        rule_status (Text, Optional): Indicates whether the rule is enabled or disabled.
        description (Text, Optional): A description of the rule.
        role_arn (Text, Optional): The Amazon Resource Name (ARN) of the IAM role associated with the rule.
            If you're setting an event bus in another account as the target and that account granted permission to your
            account through an organization instead of directly by the account ID, you must specify a RoleArn with proper
            permissions in the Target structure, instead of here in this parameter.
        event_bus_name (Text, Optional) : The name or ARN of the event bus to associate with this rule.
            If you omit this, the default event bus is used.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the event.
            Each tag consists of a key name and an associated value. Defaults to None.
            * Key (str, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * Value(str, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        [cloudwatch-events-rule-id]:
          aws.events.rule.present:
            - name: event_rule
            - arn: arn:aws:events:us-west-1:000000000000:rule/qqqqqqq
            - rule_status: ENABLED | DISABLED
            - schedule_expression: rate(5 minutes)
            - event_bus_name: default
            - tags:
              - Key: 'string'
                Value: 'string'


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            event_rule_id:
                aws.events.rule.present:
                  - name: event_rule
                  - arn: arn:aws:events:us-west-1:000000000000:rule/test
                  - rule_status: ENABLED | DISABLED
                  - schedule_expression: rate(5 minutes)
                  - event_bus_name: default
                  - tags:
                    - Key: idem_test_event_rule1
                      Value: test value

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    are_tags_updated = False
    is_role_updated = False
    existing_tags = {}
    plan_state = {}
    rule_arn = None
    before = await hub.exec.boto3.client.events.describe_rule(ctx, Name=resource_id)
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if before["result"] and before["ret"]:
        result["comment"] = (f"'{name}' already exists",)
        rule_arn = before["ret"]["Arn"]
        tags_return = await hub.exec.boto3.client.events.list_tags_for_resource(
            ctx, ResourceARN=rule_arn
        )
        result["result"] = tags_return.get("result")
        if result["result"]:
            existing_tags = tags_return.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags_return["comment"]
            return result

        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=before["ret"],
            idem_resource_name=name,
            tags=existing_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["old_state"] = resource_converted["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        update_return = await hub.exec.aws.events.rule.update_events_rule(
            ctx=ctx,
            old_targets=result["old_state"].get("targets"),
            new_targets=targets,
            schedule_expression=schedule_expression,
            event_pattern=event_pattern,
            state=rule_status,
            plan_state=plan_state,
            role_arn=role_arn,
            description=description,
            event_bus_name=event_bus_name,
            resource_id=resource_id,
        )
        result["result"] = result["result"] and update_return["result"]
        result["comment"] = result["comment"] + update_return["comment"]
        if result["result"]:
            is_role_updated = True

        if tags is not None and tags != result["old_state"].get("tag"):
            # Update tags
            update_tags_result = await hub.exec.aws.events.tag.update_tags(
                ctx=ctx,
                resource_id=rule_arn,
                old_tags=result["old_state"].get("tag"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_result["comment"]
            result["result"] = result["result"] and update_tags_result["result"]

            if not result["result"]:
                return result
            are_tags_updated = result["result"]

            if ctx.get("test", False) and (update_tags_result["ret"] is not None):
                plan_state["tags"] = update_tags_result["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "targets": targets,
                    "tags": tags,
                    "resource_id": name,
                    "schedule_expression": schedule_expression,
                    "event_pattern": event_pattern,
                    "description": description,
                    "event_bus_name": event_bus_name,
                    "role_arn": role_arn,
                    "rule_status": rule_status,
                },
            )
            result["comment"] = (f"Would create aws.events.rule '{name}'",)
            return result
        try:
            ret = await hub.exec.boto3.client.events.put_rule(
                ctx=ctx,
                Name=name,
                ScheduleExpression=schedule_expression,
                EventPattern=event_pattern,
                State=rule_status,
                Description=description,
                RoleArn=role_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
                EventBusName=event_bus_name,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            if targets is not None:
                target_ret = await hub.exec.boto3.client.events.put_targets(
                    ctx=ctx,
                    Rule=name,
                    Targets=targets,
                    EventBusName=event_bus_name,
                )
                result["result"] = result["result"] and target_ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + ret["comment"]
                    return result
            resource_id = name
            rule_arn = ret["ret"]["RuleArn"]
            existing_tags = (
                hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags) if tags else []
            )
            result["comment"] = (f"Created aws.events.rule '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"]) or is_role_updated:
        after = await hub.exec.boto3.client.events.describe_rule(ctx, Name=resource_id)
        final_updated_tags = existing_tags
        if are_tags_updated:
            tags = await hub.exec.boto3.client.events.list_tags_for_resource(
                ctx, ResourceARN=rule_arn
            )
            if tags.get("result") and tags.get("ret"):
                final_updated_tags = tags.get("ret").get("Tags")
        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=after["ret"],
            idem_resource_name=name,
            tags=final_updated_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["new_state"] = resource_converted["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
    event_bus_name: str = None,
) -> Dict[str, Any]:
    r"""

    Deletes the specified rule. Before you can delete the rule, you must remove all targets, using RemoveTargets.
    When you delete a rule, incoming events might continue to match to the deleted rule. Allow a short period of
    time for changes to take effect. If you call delete rule multiple times for the same rule, all calls will
    succeed. When you call delete rule for a non-existent custom eventbus, ResourceNotFoundException is returned.
    Managed rules are rules created and managed by another Amazon Web Services service on your behalf. These rules
    are created by those other Amazon Web Services to support functionality in those services. You can
    delete these rules using the Force option, but you should do so only if you are sure the other service is not
    still using that rule.

    Args:
        ctx:
        hub:
        resource_id(Text): The name of the AWS CloudWatch Events Rule.
        name(Text): A name, ID to identify the resource.
        event_bus_name (Text, Optional) : The name or ARN of the event bus to associate with this rule.
            If you omit this, the default event bus is used.

    Request Syntax:
    [rule-id]:
      aws.events.rule.absent:
      - name: 'string'
      - resource_id: 'string'
      - event_bus_name: 'string'
      - qualifier: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.events.rule.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.events.describe_rule(ctx, Name=resource_id)
    if not before["result"]:
        result["comment"] = (f"aws.events.rule '{name}' already absent",)
    else:
        tags = await hub.exec.boto3.client.events.list_tags_for_resource(
            ctx, ResourceARN=before["ret"].get("Arn")
        )
        result["result"] = tags.get("result")
        if result["result"]:
            existing_tags = tags.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags["comment"]
            return result

        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=before["ret"],
            idem_resource_name=before["ret"].get("Name"),
            tags=existing_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["old_state"] = resource_converted["ret"]
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.events.rule '{name}'",
            )
            return result
        else:
            try:
                targets = await hub.exec.boto3.client.events.list_targets_by_rule(
                    ctx, Rule=name
                )
                result["result"] = result["result"] and targets["result"]
                if result["result"] and targets["ret"]["Targets"]:
                    ids = []
                    for target in targets["ret"]["Targets"]:
                        ids.append(target.get("Id"))
                    remove_targets = await hub.exec.boto3.client.events.remove_targets(
                        ctx, Rule=name, Ids=ids
                    )
                    result["result"] = result["result"] and remove_targets["result"]
                    if not result["result"]:
                        result["comment"] = (
                            result["comment"] + remove_targets["comment"]
                        )
                        return result
                else:
                    result["comment"] = result["comment"] + targets["comment"]
                    return result
                ret = await hub.exec.boto3.client.events.delete_rule(
                    ctx, Name=name, EventBusName=event_bus_name
                )
                result["result"] = ret["result"] and result["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + ret["comment"]
                    return result
                result["comment"] = result["comment"] + (
                    f"Deleted aws.events.rule '{name}'",
                )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists your Amazon EventBridge rules. You can either list all the rules or you can provide a prefix to match to
    the rule names. ListRules does not list the targets of a rule. To see the targets associated with a rule, use
    ListTargetsByRule.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.events.rule
    """

    result = {}
    ret = await hub.exec.boto3.client.events.list_rules(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.events.rule {ret['comment']}")
        return result
    for rule in ret["ret"]["Rules"]:
        tags = await hub.exec.boto3.client.events.list_tags_for_resource(
            ctx, ResourceARN=rule.get("Arn")
        )
        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=rule,
            idem_resource_name=rule.get("Name"),
            tags=tags.get("ret").get("Tags") if tags.get("result") else None,
        )
        if not resource_converted["result"]:
            hub.log.warning(
                f"Could not describe aws.events.rule '{rule.get('Name')}' "
                f"with error {resource_converted['comment']}"
            )
        else:
            result[rule.get("Arn")] = {
                "aws.events.rule.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_converted[
                        "ret"
                    ].items()
                ]
            }
    return result



"""

hub.exec.boto3.client.wafv2.associate_web_acl
hub.exec.boto3.client.wafv2.disassociate_web_acl
hub.exec.boto3.client.wafv2.get_web_acl_for_resource
hub.exec.boto3.client.wafv2.list_web_acls
"""
import asyncio
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    web_acl_arn: str,
    resource_arn: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""

    Associates a web ACL with a regional application resource, to protect the resource.
    A regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API, or an AppSync GraphQL API.

    Args:
        name(Text): The idem name for the resource
        resource_id(Text, Optional): The identifier for this object. resource_arn in this case
        web_acl_arn: The Amazon Resource Name (ARN) of the web ACL that you want to associate with the resource.
        resource_arn: The Amazon Resource Name (ARN) of the resource to associate with the web ACL.

    Request Syntax:
        [resource-arn]:
          aws.wafv2.associate_web_acl.present:
          - name: 'string'
          - resource_id: 'string'
          - web_acl_arn: 'string'
          - resource_arn: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/waf-cli-alb/1ea17125f8b25a2a:
              aws.wafv2.associate_web_acl.present:
              - name: arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/waf-cli-alb/1ea17125f8b25a2a
              - web_acl_arn: arn:aws:wafv2:us-west-2:123456789012:regional/webacl/test-cli/a1b2c3d4-5678-90ab-cdef-EXAMPLE11111
              - resource_arn: arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/waf-cli-alb/1ea17125f8b25a2a
              - resource_id: arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/waf-cli-alb/1ea17125f8b25a2a

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.wafv2.get_web_acl_for_resource(
        ctx, ResourceArn=resource_arn
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if before["ret"].get("WebACL", None):
        result[
            "old_state"
        ] = hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_resource_association_to_present(
            web_acl_arn=web_acl_arn, resource_arn=resource_arn, idem_resource_name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (f"aws.waf.web_acl_association '{name}' already exists",)
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "web_acl_arn": web_acl_arn,
                    "resource_arn": resource_arn,
                },
            )
            result["comment"] = (f"Would create aws.waf.web_acl_association '{name}'",)
            return result
        for i in range(5):
            # Handling association to a newly created web ACL
            ret = await hub.exec.boto3.client.wafv2.associate_web_acl(
                ctx,
                WebACLArn=web_acl_arn,
                ResourceArn=resource_arn,
            )
            if ret["result"]:
                break
            else:
                # Waiting for web_acl to be in consumable state
                if "WAFUnavailableEntityException" in str(ret["comment"]):
                    await asyncio.sleep(60)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + ret["comment"]
            return result
        result["comment"] = result["comment"] + (
            f"Created aws.waf.web_acl_association '{name}'",
        )
        result[
            "new_state"
        ] = hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_resource_association_to_present(
            web_acl_arn=web_acl_arn, resource_arn=resource_arn, idem_resource_name=name
        )
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Disassociates a web ACL from a regional application resource.
    A regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API, or an AppSync GraphQL API.

    Args:
        ctx:
        hub:
        name(Text): The idem name for the resource
        resource_id(Text, Optional): The identifier for this object. resource_arn in this case

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            AWSServiceRoleForEC2Spot:
              aws.wafv2.associate_web_acl.absent:
              - name: arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/waf-cli-alb/1ea17125f8b25a2a
              - resource_id: arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/waf-cli-alb/1ea17125f8b25a2a

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.wafv2.associate_web_acl", name=name
        )
        return result
    before = await hub.exec.boto3.client.wafv2.get_web_acl_for_resource(
        ctx, ResourceArn=resource_id
    )
    if not before["result"]:
        if "Invalid type" in str(before["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.wafv2.associate_web_acl", name=name
            )
        else:
            result["result"] = False
            result["comment"] = before["comment"]
        return result
    if before["ret"]:
        before = before["ret"].get("WebACL", None)
    if before is None:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.wafv2.associate_web_acl", name=name
        )
    else:
        web_acl_arn = before["ARN"]
        result[
            "old_state"
        ] = hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_resource_association_to_present(
            web_acl_arn=web_acl_arn, resource_arn=resource_id, idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = (
                f"Would disassociate aws.waf.web_acl_association '{name}'",
            )
            return result
        else:
            ret = await hub.exec.boto3.client.wafv2.disassociate_web_acl(
                ctx,
                ResourceArn=resource_id,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted aws.waf.web_acl_association '{name}'",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Retrieves the Amazon Resource Names (ARNs) for the regional resources that are associated with the specified web ACL


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.wafv2.associate_web_acl
    """

    result = {}
    web_acls = await hub.exec.boto3.client.wafv2.list_web_acls(ctx, Scope="REGIONAL")
    resource_types = ["APPLICATION_LOAD_BALANCER", "API_GATEWAY", "APPSYNC"]
    if not web_acls["result"]:
        hub.log.debug(f"Could not retrieve list_web_acls {web_acls['comment']}")
        return {}

    web_acl_list = web_acls["ret"]["WebACLs"]
    for web_acl in web_acl_list:
        web_acl_arn = web_acl["ARN"]
        for resource_type in resource_types:
            resource_for_web_acl = (
                await hub.exec.boto3.client.wafv2.list_resources_for_web_acl(
                    ctx, WebACLArn=web_acl_arn, ResourceType=resource_type
                )
            )
            if not resource_for_web_acl["result"]:
                hub.log.debug(
                    f"Could not retrieve resource for web_acl {web_acls['comment']}"
                )
                continue
            for resource_arn in resource_for_web_acl["ret"]["ResourceArns"]:
                translated_resource = hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_resource_association_to_present(
                    web_acl_arn=web_acl_arn,
                    resource_arn=resource_arn,
                    idem_resource_name=resource_arn,
                )
                result[translated_resource["resource_id"]] = {
                    "aws.wafv2.associate_web_acl.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in translated_resource.items()
                    ]
                }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.dynamodb.create_table
hub.exec.boto3.client.dynamodb.delete_table
hub.exec.boto3.client.dynamodb.describe_table
hub.exec.boto3.client.dynamodb.list_tables
hub.exec.boto3.client.dynamodb.update_table
resource = hub.tool.boto3.resource.create(ctx, "dynamodb", "Table", name)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete_item, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, get_item, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, put_item, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, query, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, scan, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, update, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, update_item, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


async def present(
    hub,
    ctx,
    name: str,
    attribute_definitions: List[
        make_dataclass(
            "AttributeDefinition", [("AttributeName", str), ("AttributeType", str)]
        )
    ],
    key_schema: List[
        make_dataclass("KeySchemaElement", [("AttributeName", str), ("KeyType", str)])
    ],
    resource_id: str = None,
    billing_mode: str = None,
    sse_specification: make_dataclass(
        "SSESpecification",
        [
            ("Enabled", bool, field(default=None)),
            ("SSEType", str, field(default=None)),
            ("KMSMasterKeyId", str, field(default=None)),
        ],
    ) = None,
    stream_specification: make_dataclass(
        "StreamSpecification",
        [("StreamEnabled", bool), ("StreamViewType", str, field(default=None))],
    ) = None,
    provisioned_throughput: make_dataclass(
        "ProvisionedThroughput",
        [("ReadCapacityUnits", int), ("WriteCapacityUnits", int)],
    ) = None,
    local_secondary_indexes: List[
        make_dataclass(
            "LocalSecondaryIndex",
            [
                ("IndexName", str),
                (
                    "KeySchemaElement",
                    List[
                        make_dataclass(
                            "KeySchema", [("AttributeName", str), ("KeyType", str)]
                        )
                    ],
                ),
                (
                    "Projection",
                    make_dataclass(
                        "Projection",
                        [
                            ("ProjectionType", str, field(default=None)),
                            ("NonKeyAttributes", List[str], field(default=None)),
                        ],
                    ),
                ),
            ],
        )
    ] = None,
    global_secondary_indexes: List[
        make_dataclass(
            "GlobalSecondaryIndex",
            [
                ("IndexName", str),
                (
                    "KeySchemaElement",
                    List[
                        make_dataclass(
                            "KeySchema", [("AttributeName", str), ("KeyType", str)]
                        )
                    ],
                ),
                (
                    "Projection",
                    make_dataclass(
                        "Projection",
                        [
                            ("ProjectionType", str, field(default=None)),
                            ("NonKeyAttributes", List[str], field(default=None)),
                        ],
                    ),
                ),
                (
                    "ProvisionedThroughput",
                    make_dataclass(
                        "ProvisionedThroughput",
                        [("ReadCapacityUnits", int), ("WriteCapacityUnits", int)],
                    ),
                    field(default=None),
                ),
            ],
        )
    ] = None,
    replica_updates: List = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    table_class: str = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    The CreateTable operation adds a new table to your account. In an Amazon Web Services account, table names must
    be unique within each Region. That is, you can have two tables with same name if you create the tables in
    different Regions.  CreateTable is an asynchronous operation. Upon receiving a CreateTable request, DynamoDB
    immediately returns a response with a TableStatus of CREATING. After the table is created, DynamoDB sets the
    TableStatus to ACTIVE. You can perform read and write operations only on an ACTIVE table.  You can optionally
    define secondary indexes on the new table, as part of the CreateTable operation. If you want to create multiple
    tables with secondary indexes on them, you must create the tables sequentially. Only one table with secondary
    indexes can be in the CREATING state at any given time. You can use the DescribeTable action to check the table
    status.

    Args:
        name(Text): A name of the table to create.
        resource_id(Text, optional): AWS DynamoDB table name.
        key_schema(List[Dict[str, Any]]): Specifies the attributes that make up the primary key for a table or an index. The attributes in
            KeySchema must also be defined in the AttributeDefinitions array. For more information, see Data
            Model in the Amazon DynamoDB Developer Guide.
            Each KeySchemaElement in the array is composed of:
                AttributeName - The name of this key attribute.
                KeyType - The role that the key attribute will assume:
                    HASH - partition key
                    RANGE - sort key
            The partition key of an item is also known as its hash attribute. The term "hash attribute" derives from the DynamoDB usage of
            an internal hash function to evenly distribute data items across partitions, based on their
            partition key values. The sort key of an item is also known as its range attribute. The term
            "range attribute" derives from the way DynamoDB stores items with the same partition key
            physically close together, in sorted order by the sort key value.  For a simple primary key
            (partition key), you must provide exactly one element with a KeyType of HASH. For a composite
            primary key (partition key and sort key), you must provide exactly two elements, in this order:
            The first element must have a KeyType of HASH, and the second element must have a KeyType of
            RANGE. For more information, see Working with Tables in the Amazon DynamoDB Developer Guide.
            * AttributeName (str): The name of a key attribute.
            * KeyType (str): The role that this key attribute will assume:
                    HASH - partition key
                    RANGE - sort key
                The partition key of an item is also known as its hash attribute. The term "hash attribute" derives
                from DynamoDB's usage of an internal hash function to evenly distribute data items across
                partitions, based on their partition key values. The sort key of an item is also known as its
                range attribute. The term "range attribute" derives from the way DynamoDB stores items with the
                same partition key physically close together, in sorted order by the sort key value.
        attribute_definitions(List[Dict[str, Any]]): An array of attributes that describe the key schema for the table and indexes.
            * AttributeName (str): A name for the attribute.
            * AttributeType (str): The data type for the attribute, where:
                S - the attribute is of type String
                N - the attribute is of type Number
                B - the attribute is of type Binary
        replica_updates: A list of replica update actions (create, delete, or update) for the table.
        table_class(Text, optional): The table class of the new table. Valid values are STANDARD and
                STANDARD_INFREQUENT_ACCESS.
        billing_mode(Text, optional): Controls how users are charged for read and write throughput and how they manage
            capacity.
                PROVISIONED - AWS recommends using PROVISIONED for predictable workloads. PROVISIONED sets the billing
                            mode to Provisioned Mode.
                PAY_PER_REQUEST - AWS recommends using PAY_PER_REQUEST for unpredictable workloads. PAY_PER_REQUEST
                            sets the billing mode to On-Demand Mode .
        global_secondary_indexes(List, optional): One or more global secondary indexes (the maximum is 20) to be created on the table.
            * IndexName (str): The name of the global secondary index. The name must be unique among all other indexes on this
                table.
            * KeySchema (List[Dict[str, Any]]): The complete key schema for a global secondary index
                * AttributeName (str): The name of a key attribute.
                * KeyType (str): The role that this key attribute will assume:
                        HASH - partition key
                        RANGE - sort key
                    The partition key of an item is also known as its hash attribute. The term "hash attribute" derives
                    from DynamoDB's usage of an internal hash function to evenly distribute data items across
                    partitions, based on their partition key values. The sort key of an item is also known as its
                    range attribute. The term "range attribute" derives from the way DynamoDB stores items with the
                    same partition key physically close together, in sorted order by the sort key value.
            * Projection (Dict[str, Any]): Represents attributes that are copied (projected) from the table into the global secondary
                index. These are in addition to the primary key attributes and index key attributes, which are
                automatically projected.
                * ProjectionType (str, optional): The set of attributes that are projected into the index:
                        KEYS_ONLY - Only the index and primary keys are projected into the index.
                        INCLUDE - In addition to the attributes described in KEYS_ONLY, the secondary index will include other non-key attributes that you specify.
                        ALL - All of the table attributes are projected into the index.
            * ProvisionedThroughput (Dict[str, Any], optional): Represents the provisioned throughput settings for the specified global secondary index. For
                current minimum and maximum provisioned throughput values, see Service, Account, and Table
                Quotas in the Amazon DynamoDB Developer Guide.
                * ReadCapacityUnits (int): The maximum number of strongly consistent reads consumed per second before DynamoDB returns a
                    ThrottlingException. For more information, see Specifying Read and Write Requirements in the
                    Amazon DynamoDB Developer Guide. If read/write capacity mode is PAY_PER_REQUEST the value is set
                    to 0.
                * WriteCapacityUnits (int): The maximum number of writes consumed per second before DynamoDB returns a ThrottlingException.
                    For more information, see Specifying Read and Write Requirements in the Amazon DynamoDB
                    Developer Guide. If read/write capacity mode is PAY_PER_REQUEST the value is set to 0.
        local_secondary_indexes(List[Dict[str, Any]], optional): One or more local secondary indexes (the maximum is 5) to be created on the table. Each index is
            scoped to a given partition key value. There is a 10 GB size limit per partition key value;
            otherwise, the size of a local secondary index is unconstrained.
            * IndexName (str): The name of the local secondary index. The name must be unique among all other indexes on this
                table.
            * KeySchema (List[Dict[str, Any]]): The complete key schema for the local secondary index.
                * AttributeName (str): The name of a key attribute.
                * KeyType (str): The role that this key attribute will assume:
                        HASH - partition key
                        RANGE - sort key
                    The partition key of an item is also known as its hash attribute. The term "hash attribute" derives
                    from DynamoDB's usage of an internal hash function to evenly distribute data items across
                    partitions, based on their partition key values. The sort key of an item is also known as its
                    range attribute. The term "range attribute" derives from the way DynamoDB stores items with the
                    same partition key physically close together, in sorted order by the sort key value.
            * Projection (Dict[str, Any]): Represents attributes that are copied (projected) from the table into the local secondary index.
                These are in addition to the primary key attributes and index key attributes, which are
                automatically projected.
                * ProjectionType (str, optional): The set of attributes that are projected into the index:
                        KEYS_ONLY - Only the index and primary keys are projected into the index.
                        INCLUDE - In addition to the attributes described in KEYS_ONLY, the secondary index will include other non-key attributes that you specify.
                        ALL - All of the table attributes are projected into the index.
                * NonKeyAttributes (List[str], optional): Represents the non-key attribute names which will be projected into the index. For local
                    secondary indexes, the total count of NonKeyAttributes summed across all of the local secondary
                    indexes, must not exceed 100. If you project the same attribute into two different indexes, this
                    counts as two distinct attributes when determining the total.
        provisioned_throughput(Dict, optional): Represents the provisioned throughput settings for a specified table or index. The settings can
            be modified using the UpdateTable operation.  If you set BillingMode as PROVISIONED, you must
            specify this property. If you set BillingMode as PAY_PER_REQUEST, you cannot specify this
            property. For current minimum and maximum provisioned throughput values, see Service, Account,
            and Table Quotas in the Amazon DynamoDB Developer Guide. Defaults to None.
            * (ReadCapacityUnits): The maximum number of strongly consistent reads consumed per second before DynamoDB returns a
                ThrottlingException. For more information, see Specifying Read and Write Requirements in the
                Amazon DynamoDB Developer Guide. If read/write capacity mode is PAY_PER_REQUEST the value is set
                to 0.
            * (WriteCapacityUnits): The maximum number of writes consumed per second before DynamoDB returns a ThrottlingException.
                For more information, see Specifying Read and Write Requirements in the Amazon DynamoDB
                Developer Guide. If read/write capacity mode is PAY_PER_REQUEST the value is set to 0.
        stream_specification(Dict[str, Any], optional): The settings for DynamoDB Streams on the table. These settings consist of:
            StreamEnabled - Indicates whether DynamoDB Streams is to be enabled (true) or disabled (false).
            StreamViewType - When an item in the table is modified, StreamViewType determines what
                information is written to the table's stream. Valid values for StreamViewType are:
                    KEYS_ONLY - Only the key attributes of the modified item are written to the stream.
                    NEW_IMAGE - The entire item, as it appears after it was modified, is written to the stream.
                    OLD_IMAGE - The entire item, as it appeared before it was modified, is written to the stream.
                    NEW_AND_OLD_IMAGES - Both the new and the old item images of the item are written to the stream. Defaults to None.
            * StreamEnabled (bool): Indicates whether DynamoDB Streams is enabled (true) or disabled (false) on the table.
            * StreamViewType (str, optional):  When an item in the table is modified, StreamViewType determines what information is written to
                the stream for this table. Valid values for StreamViewType are:
                    KEYS_ONLY - Only the key attributes of the modified item are written to the stream.
                    NEW_IMAGE - The entire item, as it appears after it was modified, is written to the stream.
                    OLD_IMAGE - The entire item, as it appeared before it was modified, is written to the stream.
                    NEW_AND_OLD_IMAGES - Both the new and the old item images of the item are written to the stream.
        sse_specification(Dict[str, Any], optional): Represents the settings used to enable server-side encryption. Defaults to None.
            * Enabled (bool, optional): Indicates whether server-side encryption is done using an Amazon Web Services managed key or an
                Amazon Web Services owned key. If enabled (true), server-side encryption type is set to KMS and
                an Amazon Web Services managed key is used (KMS charges apply). If disabled (false) or not
                specified, server-side encryption is set to Amazon Web Services owned key.
            * SSEType (str, optional): Server-side encryption type.
                The only supported value is:
                    KMS - Server-side encryption that uses Key Management Service. The key is stored in your account
                        and is managed by KMS (KMS charges apply).
            * KMSMasterKeyId (str, optional): The KMS key that should be used for the KMS encryption. To specify a key, use its key ID, Amazon
                Resource Name (ARN), alias name, or alias ARN. Note that you should only provide this parameter
                if the key is different from the default DynamoDB key alias/aws/dynamodb.
        tags(List or Dict, optional): The tags to assign to the route_table. Defaults to None. Format can be [{"Key": tag-key, "Value": tag-value}] or dict {tag-key: tag-value}
            * Key (str): The key of the tag. Tag keys are case sensitive. Each DynamoDB table can only have up to one tag
                with the same key. If you try to add an existing tag (same key), the existing tag value will be
                updated to the new value.
            * Value (str): The value of the tag. Tag values are case-sensitive and can be null.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS IAM Policy.
            * create (Dict): Timeout configuration for creating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
            * update(Dict, optional): Timeout configuration for updating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [dynamodb_table-name]:
          aws.dynamodb.table.present:
          - name: 'string'
          - attribute_definitions: 'list'
          - key_schema: 'list'
          - resource_id: 'string'
          - billing_mode: 'string'
          - sse_specification: 'dict'
          - stream_specification: 'dict'
          - provisioned_throughput: 'dict'
          - local_secondary_indexes: 'list'
          - global_secondary_indexes: 'list'
          - tags:
            - Key: 'string'
              Value: 'string'
          - table_class: 'str'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_dyanmodb_table:
              aws.dynamodb.table.present:
                - name: test_dyanmodb_table
                - resource_id: dyanmodb_table-1e8965drf2c56902g
                - attribute_definitions:
                  - AttributeName: Artist
                    AttributeType: S
                  - AttributeName: SongTitle
                    AttributeType: S
                - key_schema:
                  - AttributeName: Artist
                    AttributeType: HASH
                  - AttributeName: SongTitle
                    AttributeType: RANGE
                - provisioned_throughput = { "ReadCapacityUnits": 123, "WriteCapacityUnits": 123, }
                - tags:
                  - Key: Name
                    Value: test-nat-gateway
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    is_table_updated = False
    existing_tags = None
    before = None
    plan_state = {}
    are_tags_updated = False
    table_arn = None
    if resource_id:
        before = await hub.exec.boto3.client.dynamodb.describe_table(
            ctx, TableName=resource_id
        )
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before and before["result"]:
        result["comment"] = (f"aws.dynamodb.table '{name}' is already present",)
        table_resource = before["ret"].get("Table")
        table_arn = table_resource.get("TableArn")
        tags_ret = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
            ctx, ResourceArn=table_arn
        )
        result["result"] = tags_ret.get("result")
        if result["result"]:
            existing_tags = tags_ret.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + (tags_ret["comment"],)
            return result
        result[
            "old_state"
        ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
            raw_resource=before["ret"].get("Table"),
            idem_resource_name=name,
            tags=existing_tags,
        )
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = (
            await hub.exec.aws.dynamodb.table.compare_inputs_and_update_dynamodb_table(
                ctx=ctx,
                plan_state=plan_state,
                attribute_definitions=attribute_definitions,
                name=name,
                billing_mode=billing_mode,
                provisioned_throughput=provisioned_throughput,
                global_secondary_indexes=global_secondary_indexes,
                sse_specification=sse_specification,
                stream_specification=stream_specification,
                replica_updates=replica_updates,
                table_class=table_class,
                timeout=timeout,
            )
        )
        result["result"] = result["result"] and update_ret["result"]
        result["comment"] = result["comment"] + update_ret["comment"]
        if result["result"]:
            is_table_updated = True
        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_tags_result = await hub.exec.aws.dynamodb.tag.update_tags(
                ctx=ctx,
                resource_arn=table_arn,
                old_tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(existing_tags)
                if existing_tags
                else {},
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_result["comment"]
            result["result"] = result["result"] and update_tags_result["result"]

            if not result["result"]:
                return result
            are_tags_updated = result["result"]

            if ctx.get("test", False) and (update_tags_result["ret"] is not None):
                plan_state["tags"] = update_tags_result["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "attribute_definitions": attribute_definitions,
                    "key_schema": key_schema,
                    "name": name,
                    "stream_specification": stream_specification,
                    "billing_mode": billing_mode,
                    "table_class": table_class,
                    "local_secondary_indexes": local_secondary_indexes,
                    "global_secondary_indexes": global_secondary_indexes,
                    "sse_specification": sse_specification,
                    "provisioned_throughput": provisioned_throughput,
                    "replica_updates": replica_updates,
                    "tags": tags,
                    "resource_id": name,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.dynamodb.table", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.dynamodb.create_table(
                ctx,
                AttributeDefinitions=attribute_definitions,
                KeySchema=key_schema,
                TableName=name,
                BillingMode=billing_mode,
                GlobalSecondaryIndexes=global_secondary_indexes,
                LocalSecondaryIndexes=local_secondary_indexes,
                ProvisionedThroughput=provisioned_throughput,
                SSESpecification=sse_specification,
                StreamSpecification=stream_specification,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags is not None
                else None,
                TableClass=table_class,
            )
            result["result"] = ret["result"]
            if not ret["result"]:
                result["comment"] = ret["comment"]
                return result
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            hub.log.debug(f"Waiting on creating aws.dynamodb.table '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "dynamodb",
                    "table_exists",
                    TableName=name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            resource_id = ret["ret"]["TableDescription"]["TableName"]
            table_arn = ret["ret"].get("TableDescription").get("TableArn")
            existing_tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            result["comment"] = result["comment"] + (
                f"Created aws.dynamodb.table '{resource_id}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"]) or is_table_updated:
        after = await hub.exec.boto3.client.dynamodb.describe_table(
            ctx, TableName=resource_id
        )
        final_updated_tags = existing_tags
        if are_tags_updated:
            tags_ret = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
                ctx, ResourceArn=table_arn
            )
            if tags_ret.get("result") and tags_ret.get("ret"):
                final_updated_tags = tags_ret.get("ret").get("Tags")
        result[
            "new_state"
        ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
            raw_resource=after["ret"].get("Table"),
            idem_resource_name=name,
            tags=final_updated_tags,
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    The DeleteTable operation deletes a table and all of its items. After a DeleteTable request, the specified table
    is in the DELETING state until DynamoDB completes the deletion. If the table is in the ACTIVE state, you can
    delete it. If a table is in CREATING or UPDATING states, then DynamoDB returns a ResourceInUseException. If the
    specified table does not exist, DynamoDB returns a ResourceNotFoundException. If table is already in the
    DELETING state, no error is returned.   DynamoDB might continue to accept data read and write operations, such
    as GetItem and PutItem, on a table in the DELETING state until the table deletion is complete.  When you delete
    a table, any indexes on that table are also deleted. If you have DynamoDB Streams enabled on the table, then the
    corresponding stream on that table goes into the DISABLED state, and the stream is automatically deleted after
    24 hours. Use the DescribeTable action to check the status of the table.

    Args:
        name(Text): The Idem name of the dynamoDB table.
        resource_id(Text, optional): AWS DynamoDB table name.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS DynamoDB Table.
            * delete (string) -- Timeout configuration for deletion of a DynamoDB Table
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [dynamodb_table-name]:
          aws.dynamodb.table.absent:
          - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_dyanmoDB-table:
              aws.dynamodb.table.absent:
                - name: test_dynamoDB-table
                - resource_id: dyanmoDB_table-1e8965drf2c56902g:
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.dynamodb.table", name=name
        )
        return result
    before = await hub.exec.boto3.client.dynamodb.describe_table(
        ctx, TableName=resource_id
    )
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.dynamodb.table", name=name
        )
    else:
        tags = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
            ctx, ResourceArn=before["ret"].get("Table").get("TableArn")
        )
        if tags.get("result"):
            existing_tags = tags.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags["comment"]
            result["result"] = False
            return result

        result[
            "old_state"
        ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
            raw_resource=before["ret"].get("Table"),
            idem_resource_name=name,
            tags=existing_tags,
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.dynamodb.table", name=name
            )
            return result
        else:
            try:
                ret = await hub.exec.boto3.client.dynamodb.delete_table(
                    ctx, TableName=resource_id
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=15,
                    default_max_attempts=40,
                    timeout_config=timeout.get("delete") if timeout else None,
                )
                hub.log.debug(f"Waiting on deleting aws.dynamodb.table '{name}'")
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "dynamodb",
                        "table_not_exists",
                        TableName=name,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
                result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.dynamodb.table", name=name
                )
                after = await hub.exec.boto3.client.dynamodb.describe_table(
                    ctx, TableName=resource_id
                )
                if after["result"] and after["ret"]:
                    result[
                        "new_state"
                    ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
                        raw_resource=after["ret"].get("Table"),
                        idem_resource_name=name,
                        tags=existing_tags,
                    )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
                result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.dynamodb.table
    """

    result = {}

    ret = await hub.exec.boto3.client.dynamodb.list_tables(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe table {ret['comment']}")
        return result
    if ret and ret["ret"] and ret["ret"].get("TableNames"):
        for table_name in ret["ret"].get("TableNames"):
            resource = await hub.exec.boto3.client.dynamodb.describe_table(
                ctx, TableName=table_name
            )
            if (
                resource.get("result")
                and resource.get("ret") is not None
                and resource.get("ret").get("Table")
            ):
                table_resource = resource.get("ret").get("Table")
            else:
                continue
            tags = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
                ctx, ResourceArn=table_resource.get("TableArn")
            )
            resource_converted = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
                raw_resource=table_resource,
                idem_resource_name=table_name,
                tags=tags.get("ret").get("Tags") if tags.get("result") else None,
            )
            result[table_name] = {
                "aws.dynamodb.table.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_converted.items()
                ]
            }
    return result



from typing import Any
from typing import Dict

import pop.contract


def client(hub, path: str, context):
    """
    Get an AWS service client based on the path and call an operation on it
    idem will automatically populate credentials from acct in the client.

    The calls mirror the namespacing of boto3.client and have the same parameters

    If a response can be paginated, the full result will be asynchronously collected and returned.

    path::

        boto3.client.[service_name].[operation] [kwargs="values"]

    Examples:
        In these examples will use the service_name of "ec2" and operation of "create_vpc"

        Call from the CLI
        .. code-block: bash

            $ idem exec boto3.client.ec2.create_vpc CidrBlock="10.0.0.0/24"

        Call from code
        .. code-block: python

            await hub.exec.boto3.client.ec2.create_vpc(ctx, CidrBlock="10.0.0.0/24")

    :param hub:
    :param path: client.[service_name].[function_name]
    :param context: None
    :return: The result of the call
    """
    c, service_name, operation = path.split(".", maxsplit=2)
    assert c == "client"

    async def _client_caller(ctx, *args, **kwargs) -> Dict[str, Any]:
        result = {"comment": (), "ret": None, "result": True}
        try:
            if operation.endswith(".search"):
                resource = operation.split(".")[0]
                ret = await hub.tool.boto3.client.search(
                    ctx, service_name, resource, *args, **kwargs
                )
                result["result"] = bool(ret)
                result["ret"] = ret
                result["comment"] = (f"Searched in {service_name}.{resource}",)
            else:
                ret: Dict[str, Any] = await hub.tool.boto3.client.exec(
                    ctx, service_name, operation, *args, **kwargs
                )
                if hasattr(ret, "keys"):
                    keys = sorted(ret.keys())
                else:
                    keys = []
                result["comment"] = tuple(keys)
                result["ret"] = ret
                result["result"] = bool(ret)
        except Exception as e:
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
        return result

    return pop.contract.ContractedAsync(
        hub,
        contracts=[],
        func=_client_caller,
        ref=path,
        name=operation,
        implicit_hub=False,
    )


def resource(hub, path: str, context):
    """
    Get an AWS service resource based on the path.
    If the path includes an operation then a resource is created with the args which runs the operation with the kwargs
    If the path does not include an operation then pass args and kwargs to the resource class and return details

    The calls mirror the namespacing of boto3.resource and have the same parameters

    Examples:
        Call from the CLI
        -----------------

        .. code-block: bash

            $ idem exec boto3.resource.ec2.Vpc.create_subnet vpc-71d00419 CidrBlock="10.0.0.0/24"

        Call from code
        --------------
        You can make the same call from within code:

        .. code-block: python

            # ret will do error catching and format the response as an ExecReturn
            ret = await hub.exec.boto3.ec2.Vpc.create_subnet("vpc-71d00419", CidrBlock="10.0.0.0/24")

        More control and better performance come from using the underlying tools directly, like so:

        .. code-block: python

            # resource will be reusable for more operation calls
            resource = hub.tool.boto3.resource.create(ctx, "Vpc", id="vpc-71d00419")

            # ret will contain the raw result of the operation call
            ret = await hub.tool.boto3.resource.exec(resource, "create_subnet", CidrBlock="10.0.0.0/24")

    :param hub:
    :param path: client.[service_name].[function_name]
    :param context: None
    :return: The data retrieved from the given resource path
    """
    r, service_name, resource_class = path.split(".", maxsplit=2)
    assert r == "resource"

    if "." in resource_class:
        resource_class, operation = resource_class.split(".")
    else:
        operation = None

    async def _resource_caller(ctx, *args, **kwargs):
        result = {"comment": (), "ret": None, "result": True}
        try:
            if operation == "id":
                # Return the first id that matches the given kwargs
                ret = await hub.tool.boto3.resource.id(
                    ctx,
                    service_name=service_name,
                    collection=f"{resource_class}s",
                    *args,
                    **kwargs,
                )
                result["comment"] = (
                    f"Searched for id in {service_name}.{resource_class}",
                )
                result["ret"] = ret
                result["result"] = bool(ret)
            elif operation:
                # There's a function call to make on this resource
                # Run the operation with the given parameters
                resource_ = hub.tool.boto3.resource.create(
                    ctx, service_name, resource_class, *args
                )
                result["ret"] = await hub.tool.boto3.resource.exec(
                    resource_, operation, **kwargs
                )
                result["comment"] = (f"{resource_.__class__.__name__}.{operation}",)
            else:
                # The path ends with the resource name describe it with the given parameters
                resource_ = hub.tool.boto3.resource.create(
                    ctx, service_name, resource_class, *args, **kwargs
                )
                # Return raw information about the object
                result["ret"] = hub.tool.boto3.resource.dict(resource_)
                result["comment"] = (resource_.__class__.__name__,)
        except Exception as e:
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

        return result

    return pop.contract.ContractedAsync(
        hub,
        contracts=[],
        func=_resource_caller,
        ref=path,
        name=resource_class,
        implicit_hub=False,
    )



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    acl: str = None,
    create_bucket_configuration: make_dataclass(
        "CreateBucketConfiguration", [("LocationConstraint", str, field(default=None))]
    ) = None,
    grant_full_control: str = None,
    grant_read: str = None,
    grant_read_acp: str = None,
    grant_write: str = None,
    grant_write_acp: str = None,
    object_lock_enabled_for_bucket: bool = None,
    object_ownership: str = None,
    object_lock_configuration: dict = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    r"""
        name (Text) -- An Idem name of the resource.
        create_bucket_configuration(Dict[str, Any], optional): The configuration information for the bucket. Defaults to None.
            LocationConstraint (str, optional): Specifies the Region where the bucket will be created. If you don't specify a Region, the bucket
                is created in the US East (N. Virginia) Region (us-east-1).
        resource_id(Text, optional): AWS S3 Bucket name
        acl (Text, optional) -- The canned ACL to apply to the bucket.
        grant_full_control (Text, optional) --
            Allows grantee the read, write, read ACP, and write ACP permissions on the bucket.
        grant_read (Text, optional) --
            Allows grantee to list the objects in the bucket.
        grant_read_acp (Text, optional) --
            Allows grantee to read the bucket ACL.
        grant_write (Text, optional) --
            Allows grantee to create new objects in the bucket.
            For the bucket and object owners of existing objects, also allows deletions and overwrites of those objects.
        grant_write_acp (Text, optional) --
            Allows grantee to write the ACL for the applicable bucket.
        object_lock_enabled_for_bucket (boolean, optional) --
            Specifies whether you want S3 Object Lock to be enabled for the new bucket.
            ObjectLockConfiguration (dict) --
        The Object Lock configuration that you want to apply to the specified bucket.
            ObjectLockEnabled (string) -- Indicates whether this bucket has an Object Lock configuration enabled. Enable ObjectLockEnabled
                                        when you apply ObjectLockConfiguration to a bucket.
            Rule (dict) -- Specifies the Object Lock rule for the specified object. Enable the this rule when you apply ObjectLockConfiguration
                            to a bucket. Bucket settings require both a mode and a period. The period can be either Days or Years but you must select one.
                            You cannot specify Days and Years at the same time.
                DefaultRetention (dict) -- The default Object Lock retention mode and period that you want to apply to new objects placed
                                    in the specified bucket. Bucket settings require both a mode and a period. The period can be either Days or Years
                                    but you must select one. You cannot specify Days and Years at the same time.
                    Mode (string) -- The default Object Lock retention mode you want to apply to new objects placed in the specified bucket.
                                    Must be used with either Days or Years .
                    Days (integer) -- The number of days that you want to specify for the default retention period. Must be used with Mode .
                    Years (integer) -- The number of years that you want to specify for the default retention period. Must be used with Mode .
        object_ownership (Text, optional) --
            The container element for object ownership for a bucket's ownership controls.
            BucketOwnerPreferred -
                Objects uploaded to the bucket change ownership to the bucket owner if
                the objects are uploaded with the bucket-owner-full-control canned ACL.
            ObjectWriter -
                The uploading account will own the object if the object is uploaded
                with the bucket-owner-full-control canned ACL.
            BucketOwnerEnforced -
                Access control lists (ACLs) are disabled and no longer affect permissions.
                The bucket owner automatically owns and has full control over every object
                in the bucket. The bucket only accepts PUT requests that don't specify an
                ACL or bucket owner full control ACLs, such as the bucket-owner-full-control
                canned ACL or an equivalent form of this ACL expressed in the XML format.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the bucket.
            Each tag consists of a key name and an associated value. Defaults to None.
            * Key (str, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * Value(str, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        [s3-resource-id]::
          aws.s3.bucket.present:
            - acl: 'private'|'public-read'|'public-read-write'|'authenticated-read'
            - bucket: string
            - create_bucket_configuration:
                LocationConstraint: string
            - grant_full_control: string
            - grant_read: string
            - grant_read_acp: string
            - grant_write: string
            - grant_write_acp: string
            - object_lock_enabled_for_bucket: True|False
            - object_lock_configuration: 'dict'
                ObjectLockEnabled: 'string'
                  Rule:
                    DefaultRetention: 'dict'
                      Mode: 'string'
                      Days: 'integer'
            - object_ownership: BucketOwnerPreferred'|'ObjectWriter'|'BucketOwnerEnforced'
            - tags:
                - Key: 'string'
                  Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_bucket-bb7bb32e9533:
              aws.s3.bucket.present:
                - name: test_bucket-bb7bb32e9533
                - acl: 'private'
                - create_bucket_configuration:
                    LocationConstraint: 'sa-east-1'
                - object_lock_enabled_for_bucket: True
                - object_lock_configuration:
                    ObjectLockEnabled: 'Enabled'
                    Rule:
                      DefaultRetention:
                        Mode: GOVERNANCE
                        Days: 1
                - object_ownership: 'BucketOwnerEnforced'
                - tags:
                    - Key: Name1
                      Value: s3-test1
                    - Key: Name2
                      Value: s3-test2


    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    tags_list = None
    tags_dict = None
    if tags is not None:
        if isinstance(tags, Dict):
            tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            tags_dict = tags
        else:
            tags_list = tags
            tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if resource_id:
        resource = hub.tool.boto3.resource.create(ctx, "s3", "Bucket", resource_id)
        before = await hub.tool.boto3.resource.describe(resource)
    if before:
        convert_ret = await hub.tool.aws.s3.conversion_utils.convert_raw_s3_to_present(
            ctx, idem_resource_name=name
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = convert_ret["ret"]

        plan_state = copy.deepcopy(result["old_state"])
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            "aws.s3.bucket", name
        )
        if (
            object_lock_enabled_for_bucket is not None
            and object_lock_enabled_for_bucket
            != result["old_state"].get("object_lock_enabled_for_bucket")
        ):
            if not ctx.get("test", False):
                update_ret = (
                    await hub.exec.boto3.client.s3.put_object_lock_configuration(
                        ctx,
                        Bucket=resource_id,
                        ObjectLockConfiguration=object_lock_configuration,
                    )
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
            else:
                plan_state[
                    "object_lock_enabled_for_bucket"
                ] = object_lock_enabled_for_bucket
                plan_state["object_lock_configuration"] = object_lock_configuration
        if (
            tags_dict is not None
            and not hub.tool.aws.state_comparison_utils.compare_dicts(
                result["old_state"].get("tags"), tags_dict
            )
        ):
            if not ctx.get("test", False):
                update_ret = await hub.exec.boto3.client.s3.put_bucket_tagging(
                    ctx,
                    Bucket=resource_id,
                    Tagging={"TagSet": tags_list},
                )
                result["comment"] = result["comment"] + (update_ret["comment"])
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
            else:
                plan_state["tags"] = tags_dict
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "acl": acl,
                    "name": name,
                    "resource_id": resource_id,
                    "create_bucket_configuration": create_bucket_configuration,
                    "grant_full_control": grant_full_control,
                    "grant_read": grant_read,
                    "grant_read_acp": grant_read_acp,
                    "grant_write": grant_write,
                    "grant_write_acp": grant_write_acp,
                    "object_lock_enabled_for_bucket": object_lock_enabled_for_bucket,
                    "object_ownership": object_ownership,
                    "tags": tags_dict,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.s3.bucket", name=name
            )
            return result
        # Build out S3 bucket creation here
        ret = await hub.exec.boto3.client.s3.create_bucket(
            ctx,
            ACL=acl,
            Bucket=name,
            CreateBucketConfiguration=create_bucket_configuration,
            GrantFullControl=grant_full_control,
            GrantRead=grant_read,
            GrantReadACP=grant_read_acp,
            GrantWrite=grant_write,
            GrantWriteACP=grant_write_acp,
            ObjectLockEnabledForBucket=object_lock_enabled_for_bucket,
            ObjectOwnership=object_ownership,
        )
        if not ret["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.s3.bucket", name=name
        )
        resource_id = name
        if tags_list is not None:
            ret = await hub.exec.boto3.client.s3.put_bucket_tagging(
                ctx,
                Bucket=name,
                Tagging={"TagSet": tags_list},
            )
            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = False
                return result

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        convert_ret = await hub.tool.aws.s3.conversion_utils.convert_raw_s3_to_present(
            ctx, idem_resource_name=resource_id
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["new_state"] = convert_ret["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    Deletes the specified s3 bucket.

    Args:
        name(Text): The Idem name of the s3 bucket.
        resource_id(Text, Optional): AWS S3 Bucket name

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            bucket-5435423646-456464:
              aws.s3.bucket.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = result[
            "comment"
        ] + hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket", name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(ctx, "s3", "Bucket", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = result[
            "comment"
        ] + hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket", name=name
        )
        return result
    else:
        convert_ret = await hub.tool.aws.s3.conversion_utils.convert_raw_s3_to_present(
            ctx, idem_resource_name=name
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = convert_ret["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.s3.bucket", name=name
            )
            return result
        ret = await hub.exec.boto3.client.s3.delete_bucket(ctx, Bucket=resource_id)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.s3.bucket", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Obtain S3 buckets under the given context for any user.
    :param hub:
    :param ctx:
    :return:
    """
    result = {}
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe S3 buckets {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:
        resource_id = bucket.get("Name")
        convert_ret = await hub.tool.aws.s3.conversion_utils.convert_raw_s3_to_present(
            ctx, idem_resource_name=resource_id
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe s3 bucket '{resource_id}' with error {convert_ret['comment']}"
            )
            continue
        translated_resource = convert_ret["ret"]
        result[resource_id] = {
            "aws.s3.bucket.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS API Gateway v2 API resources.
"""


def convert_raw_api_to_present(hub, raw_resource: Dict[str, Any]) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 API resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(Dict[str, Any]): The AWS response to convert.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "ApiEndpoint": "api_endpoint",
            "ApiGatewayManaged": "api_gateway_managed",
            "ApiId": "api_id",
            "ApiKeySelectionExpression": "api_key_selection_expression",
            "CorsConfiguration": "cors_configuration",
            "Description": "description",
            "DisableExecuteApiEndpoint": "disable_execute_api_endpoint",
            "DisableSchemaValidation": "disable_schema_validation",
            "Name": "name",
            "ProtocolType": "protocol_type",
            "RouteSelectionExpression": "route_selection_expression",
            "Tags": "tags",
            "Version": "version",
        }
    )
    resource_translated = {
        "resource_id": raw_resource.get("ApiId"),
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



from typing import Any
from typing import Dict


async def update_monitor(
    hub, ctx, before: Dict[str, Any], monitor_name: str, monitor_arn: str
):
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}
    if monitor_name and before.get("MonitorName") != monitor_name:
        update_payload["MonitorName"] = monitor_name
    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ce.update_anomaly_monitor(
                ctx=ctx, MonitorArn=monitor_arn, **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                return result
        result["ret"] = {}
        if "MonitorName" in update_payload:
            result["ret"]["monitor_name"] = update_payload["MonitorName"]
            result["comment"] = result["comment"] + (
                f"Update monitor_name: {update_payload['MonitorName']}",
            )
    return result



from typing import Any
from typing import Dict


async def update_rule(
    hub,
    ctx,
    resource_id: str,
    before: Dict[str, Any],
    source: Dict[str, Any],
    scope: Dict[str, Any],
    frequency: str = Any,
    input_parameters: str = Any,
):
    """
    Update rule of AWS Config rule

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        resource_id: AWS Config rule name
        before: Contains current configuration for the resource
        source: Dict. The rule owner, the rule identifier. Required for update.
        scope: Dict. Defines which resources can trigger an evaluation for the rule. The scope can include one or more
         resource types, a combination of one resource type.
        frequency: String. The maximum frequency with which Config runs evaluations for a rule.
        input_parameters: A string, in JSON format

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}
    update = False
    for key, value in {
        "MaximumExecutionFrequency": frequency,
        "Scope": scope,
        "InputParameters": input_parameters,
    }.items():
        if value is not None:
            if before.get(key) != value:
                update = True
            update_payload[key] = value
    if update:
        update_payload["ConfigRuleName"] = resource_id
        update_payload["Source"] = source
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.config.put_config_rule(
                ctx=ctx, ConfigRule=update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        for key, mapped_key in {
            "MaximumExecutionFrequency": "max_execution_frequency",
            "Scope": "scope",
            "InputParameters": "input_parameters",
        }.items():
            if key in update_payload:
                if (update_payload[key] is not None) and before.get(
                    key
                ) != update_payload[key]:
                    result["ret"][mapped_key] = update_payload[key]
                    result["comment"] = result["comment"] + (
                        f"Updated {mapped_key}: {update_payload[key]}",
                    )
    return result



"""

hub.exec.boto3.client.lambda.add-permission
hub.exec.boto3.client.lambda.remove-permission
hub.exec.boto3.client.lambda.get-policy

"""
import copy
import json
from typing import Any
from typing import Dict

# if no permission exists it throws below error message.
PERMISSION_DOES_NOT_EXISTS = "ResourceNotFoundException: An error occurred (ResourceNotFoundException) when calling the GetPolicy operation"


async def present(
    hub,
    ctx,
    name: str,
    action: str,
    function_name: str,
    principal: str,
    source_arn: str = None,
    source_account: str = None,
    resource_id: str = None,
    event_source_token: str = None,
    qualifier: str = None,
    revision_id: str = None,
    principal_org_id: str = None,
    function_url_auth_type: str = None,
) -> Dict[str, Any]:
    r"""

    Add specified permission to the lambda function.

    Args:
        hub:
        ctx:
        name(str): Name/Id of the statement/permission
        action(str): The action that the principal can use on the function. For example, lambda:InvokeFunction or
                lambda:GetFunction.
        function_name(str): The name of the Lambda function, version, or alias.
                        Name formats
                        Function name - my-function (name-only), my-function:v1 (with alias).
                        Function ARN - arn:aws:lambda:us-west-2:123456789012:function:my-function .
                        Partial ARN - 123456789012:function:my-function .
        principal(str): The Amazon Web Services service or account that invokes the function. If you specify a service,
                    use source_arn or source_account to limit who can invoke the function through that service.
        source_arn(str, Optional): For Amazon Web Services , the ARN of the Amazon Web Services resource that invokes the
                    function. For example, an Amazon S3 bucket or Amazon SNS topic.
        source_account(str, Optional): For Amazon S3, the ID of the account that owns the resource. Use this together with SourceArn to
                        ensure that the resource is owned by the specified account. It is possible for an Amazon S3
                        bucket to be deleted by its owner and recreated by another account.
        resource_id(str, Optional): Name/Id of the statement/permission
        event_source_token(str, Optional): For Alexa Smart Home functions, a token that must be supplied by the invoker.
        qualifier(str, Optional): Specify a version or alias to add permissions to a published version of the function.
        revision_id(str, Optional): Only update the policy if the revision ID matches the ID that's specified. Use this option to avoid modifying a policy that has changed since you last read it.
        principal_org_id(str, Optional): The identifier for your organization in Organizations. Use this to grant permissions to all the Amazon Web Services accounts under this organization.
        function_url_auth_type(str, Optional): The type of authentication that your function URL uses. Set to AWS_IAM if you want to restrict access to authenticated IAM users only. Set to NONE if you want to bypass IAM authentication to create a public endpoint.


    Request Syntax:
        [statement_id]:
          aws.lambda_aws.function_permission.present:
          - name: 'string'
            resource_id: 'string'
            action: 'string'
            function_name: 'string'
            principal:
                'string' : 'string'

    Returns:
        Dict[str, Any]

    Examples:

        . code-block:: sls

            test-001:
              aws.lambda_aws.function_permission.present:
              - action: lambda:GetAlias
              - effect: Allow
              - function_name: ecd17a181d6588e27c976cdeff501e90750b0dcafebba907cc4aab3c
              - name: test-001
              - principal:
                  AWS: AIDAX2FJ77DC2AS7BAPBU
              - resource_id: test-001


    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    # check if lambda function exists
    ret = await hub.exec.boto3.client["lambda"].get_function(
        ctx, FunctionName=function_name
    )
    if not ret["result"]:
        result["comment"] = ret["comment"]
        result["result"] = False
        return result

    # get current set of permissions
    get_policy = await hub.exec.boto3.client["lambda"].get_policy(
        ctx, FunctionName=function_name
    )

    if not get_policy["result"] and not any(
        PERMISSION_DOES_NOT_EXISTS in msg for msg in get_policy["comment"]
    ):
        result["comment"] = get_policy["comment"]
        result["result"] = False
        return result

    if get_policy["result"]:
        final_resource_id = resource_id if resource_id else name
        raw_statements = json.loads(get_policy["ret"]["Policy"])["Statement"]

        for statement in raw_statements:
            if statement["Sid"] == final_resource_id:
                result[
                    "old_state"
                ] = hub.tool.aws.lambda_aws.function_permission.convert_raw_lambda_permission_to_present(
                    raw_resource=statement,
                    function_name=function_name,
                    revision_id=get_policy["ret"]["RevisionId"],
                )
                result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                    resource_type="aws.lambda_aws.function_permission",
                    name=final_resource_id,
                )
                result["new_state"] = copy.deepcopy(result["old_state"])
                return result

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
            resource_type="aws.lambda_aws.function_permission", name=name
        )
        principal_response = {"AWS": principal}
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={},
            desired_state={
                "name": name,
                "resource_id": name,
                "action": action,
                "function_name": function_name,
                "principal": principal_response,
                "source_arn": source_arn,
                "source_account": source_account,
                "event_source_token": event_source_token,
                "qualifier": qualifier,
                "revision_id": revision_id,
                "principal_org_id": principal_org_id,
                "function_url_auth_type": function_url_auth_type,
            },
        )
        return result

    add_permission = await hub.exec.boto3.client["lambda"].add_permission(
        ctx,
        FunctionName=function_name,
        StatementId=name,
        Action=action,
        Principal=principal,
        SourceArn=source_arn,
        SourceAccount=source_account,
        EventSourceToken=event_source_token,
        Qualifier=qualifier,
        RevisionId=revision_id,
        PrincipalOrgID=principal_org_id,
        FunctionUrlAuthType=function_url_auth_type,
    )
    if not add_permission["result"]:
        result["comment"] = add_permission["comment"]
        result["result"] = False
        return result

    get_policy = await hub.exec.boto3.client["lambda"].get_policy(
        ctx, FunctionName=function_name
    )
    if not get_policy["result"]:
        result["comment"] = get_policy["comment"]
        result["result"] = False
        return result

    raw_statements = json.loads(get_policy["ret"]["Policy"])["Statement"]

    for statement in raw_statements:
        if statement["Sid"] == name:
            result[
                "new_state"
            ] = hub.tool.aws.lambda_aws.function_permission.convert_raw_lambda_permission_to_present(
                raw_resource=statement,
                function_name=function_name,
                revision_id=get_policy["ret"]["RevisionId"],
            )
            break

    result["comment"] = hub.tool.aws.comment_utils.create_comment(
        resource_type="aws.lambda_aws.function_permission", name=name
    )

    return result


async def absent(
    hub, ctx, name: str, function_name: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""

    Remove specified permission from the lambda function.

    Args:
        hub:
        ctx:
        name(str): Name/Id of the statement/permission
        function_name(str): The name of the Lambda function, version, or alias.
                        Name formats
                        Function name - my-function (name-only), my-function:v1 (with alias).
                        Function ARN - arn:aws:lambda:us-west-2:123456789012:function:my-function .
                        Partial ARN - 123456789012:function:my-function .
        resource_id(str, Optional): Name/Id of the statement/permission

    Request Syntax:
        [statement_id]:
          aws.lambda_aws.function_permission.absent:
          - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        . code-block:: sls

            test-001:
              aws.lambda_aws.function_permission.absent:
              - name: test-001

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    ret = await hub.exec.boto3.client["lambda"].get_function(
        ctx, FunctionName=function_name
    )
    if not ret["result"]:
        result["comment"] = ret["comment"]
        result["result"] = False
        return result

    get_policy = await hub.exec.boto3.client["lambda"].get_policy(
        ctx, FunctionName=function_name
    )

    if not get_policy["result"]:
        if any(PERMISSION_DOES_NOT_EXISTS in msg for msg in get_policy["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.lambda_aws.function_permission", name=resource_id
            )
        else:
            result["comment"] = ret["comment"]
            result["result"] = False
        return result

    raw_statements = json.loads(get_policy["ret"]["Policy"])["Statement"]

    list_of_statement_ids = [statement["Sid"] for statement in raw_statements]

    if name not in list_of_statement_ids:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.lambda_aws.function_permission", name=resource_id
        )
        return result

    for statement in raw_statements:
        if statement["Sid"] == name:
            result[
                "old_state"
            ] = hub.tool.aws.lambda_aws.function_permission.convert_raw_lambda_permission_to_present(
                raw_resource=statement,
                function_name=function_name,
                revision_id=get_policy["ret"]["RevisionId"],
            )
            break

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.lambda_aws.function_permission",
            name=resource_id if resource_id else name,
        )
        return result

    remove_permission = await hub.exec.boto3.client["lambda"].remove_permission(
        ctx,
        FunctionName=function_name,
        StatementId=resource_id if resource_id else name,
    )
    if not remove_permission["result"]:
        result["comment"] = remove_permission["comment"]
        result["result"] = False

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.lambda_aws.function_permission",
        name=resource_id if resource_id else name,
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Returns permissions associated with each lambda function.

    Args:
        hub:
        ctx:

    Request Syntax:
        [statement-id]:
          aws.lambda_aws.function_permission.present:
          - name: 'string'
            resource_id: 'string'
            statements:
            - action: 'string'
              effect: 'string'
              principal:
                'string'
              statement_id: 'string'


    Returns:
        Dict[str, Any]

    Examples:

        . code-block:: sls

            test-001:
              aws.lambda_aws.function_permission.present:
              - name: test-001
                resource_id: test-001
                function_name: ecd17a181d6588e27c976cdeff501e90750b0dcafebba907cc4aab3c
                action: lambda:GetAlias
                effect: Allow
                principal:
                  AWS: arn:aws:iam::537227425989:user/idem-fixture-user-18964248-75dd-4765-80b2-100901a8c74e

    """

    result = {}

    ret = await hub.exec.boto3.client["lambda"].list_functions(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe lambda functions. Error: {ret['comment']}")
        return result

    if ret and ret["result"] and ret["ret"]["Functions"]:
        for function in ret["ret"]["Functions"]:
            function_name = function.get("FunctionName")
            get_policy = await hub.exec.boto3.client["lambda"].get_policy(
                ctx, FunctionName=function_name
            )
            if not get_policy["result"]:
                if not any(
                    PERMISSION_DOES_NOT_EXISTS in msg for msg in get_policy["comment"]
                ):
                    hub.log.debug(
                        f"Could not describe lambda functions permission for {function_name}. Error: {get_policy['comment']}"
                    )
                continue

            raw_statements = json.loads(get_policy["ret"]["Policy"])["Statement"]
            for statement in raw_statements:
                permission_translated = hub.tool.aws.lambda_aws.function_permission.convert_raw_lambda_permission_to_present(
                    raw_resource=statement,
                    function_name=function_name,
                    revision_id=get_policy["ret"]["RevisionId"],
                )
                result[statement["Sid"]] = {
                    "aws.lambda_aws.function_permission.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in permission_translated.items()
                    ]
                }

    return result



import json
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_web_acl_resource_association_to_present(
    hub,
    web_acl_arn: str,
    resource_arn: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state from Associate Web ACL to present input format.

    Args:
        hub:
        web_acl_arn: The Amazon Resource Name (ARN) of the web ACL that you want to associate with the resource.
        resource_arn: The Amazon Resource Name (ARN) of the resource to associate with the web ACL.
        idem_resource_name: The idem name for the resource.

    Returns:
        Dict[str, Any]
    """
    resource_id = resource_arn
    translated_resource = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "web_acl_arn": web_acl_arn,
        "resource_arn": resource_arn,
    }
    return translated_resource


async def convert_raw_web_acl_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    scope: str = None,
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state from Web ACL to present input format.

    Args:
        hub:
        ctx:
        raw_resource: Old state of resource or existing resource details.
        idem_resource_name: Resource name.
        scope: Specifies whether this is for an Amazon CloudFront distribution or for a regional application.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)
    resource_id = raw_resource.get("Id")
    web_acl_arn = raw_resource.get("ARN")
    resource_parameters = OrderedDict(
        {
            "DefaultAction": "default_action",
            "Description": "description",
            "Rules": "rules",
            "VisibilityConfig": "visibility_config",
            "CustomResponseBodies": "custom_response_bodies",
            "CaptchaConfig": "captcha_config",
        }
    )
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "scope": scope,
        "web_acl_arn": web_acl_arn,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = json.loads(
                json.dumps(raw_resource.get(parameter_raw))
            )

    ret_tag = await hub.exec.boto3.client.wafv2.list_tags_for_resource(
        ctx, ResourceARN=web_acl_arn
    )
    result["result"] = ret_tag["result"]
    if not result["result"]:
        result["comment"] = result["comment"] + ret_tag["comment"]
        return result
    if (
        ret_tag["ret"]
        and ret_tag["ret"]["TagInfoForResource"]
        and ret_tag["ret"]["TagInfoForResource"].get("TagList")
    ):
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            ret_tag["ret"]["TagInfoForResource"].get("TagList")
        )
    result["ret"] = resource_translated

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS GUARDDUTY to present input format.
"""


async def convert_raw_detector_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    detector_describe = {}
    resource_parameters = OrderedDict(
        {"FindingPublishingFrequency": "finding_publishing_frequency"}
    )
    ret = await hub.exec.boto3.client.guardduty.list_detectors(ctx)
    for detector_id in ret["ret"]["DetectorIds"]:
        resource_id = detector_id
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    if raw_resource and raw_resource["result"] is True:
        detector_describe = raw_resource["ret"]

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource["ret"]:
            resource_translated[parameter_present] = raw_resource["ret"].get(
                parameter_raw
            )
    data_sources_new = hub.tool.aws.guardduty.detector_utils.render_data_sources(
        detector_describe
    )

    if detector_describe["Tags"]:
        resource_translated["tags"] = dict(detector_describe["Tags"])

    resource_translated["enable"] = True

    if detector_describe["DataSources"]:
        resource_translated["data_sources"] = data_sources_new
    return resource_translated



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Exec functions for AWS API Gateway v2 Authorizer resources.
"""


async def update(
    hub,
    ctx,
    api_id: str,
    resource_id: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
) -> Dict[str, Any]:
    r"""
    Updates an AWS API Gateway v2 Authorizer resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Authorizer resource identifier in Amazon Web Services.
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        resource_parameters(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "AuthorizerCredentialsArn": "authorizer_credentials_arn",
            "AuthorizerPayloadFormatVersion": "authorizer_payload_format_version",
            "AuthorizerResultTtlInSeconds": "authorizer_result_ttl_in_seconds",
            "AuthorizerType": "authorizer_type",
            "AuthorizerUri": "authorizer_uri",
            "EnableSimpleResponses": "enable_simple_responses",
            "IdentitySource": "identity_source",
            "JwtConfiguration": "jwt_configuration",
        }
    )

    parameters_to_update = {}

    identity_source = resource_parameters.get("IdentitySource")
    if identity_source:
        if not hub.tool.aws.state_comparison_utils.are_lists_identical(
            identity_source,
            raw_resource.get("IdentitySource"),
        ):
            parameters_to_update["IdentitySource"] = identity_source

        resource_parameters.pop("IdentitySource")

    jwt_configuration = resource_parameters.get("JwtConfiguration")
    if jwt_configuration is not None:
        update_jwt_configuration = False
        old_jwt_configuration = raw_resource.get("JwtConfiguration")

        if not hub.tool.aws.state_comparison_utils.are_lists_identical(
            jwt_configuration.get("Audience"),
            old_jwt_configuration.get("Audience"),
        ):
            update_jwt_configuration = True
        elif jwt_configuration.get("Issuer") is not None and jwt_configuration.get(
            "Issuer"
        ) != old_jwt_configuration.get("Issuer"):
            update_jwt_configuration = True

        if update_jwt_configuration:
            parameters_to_update["JwtConfiguration"] = jwt_configuration

        resource_parameters.pop("JwtConfiguration")

    for key, value in resource_parameters.items():
        if value is not None and value != raw_resource.get(key):
            parameters_to_update[key] = resource_parameters[key]

    if parameters_to_update:
        result["ret"] = {}
        for parameter_raw, parameter_present in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

        if ctx.get("test", False):
            result["comment"] = (
                f"Would update parameters: " + ",".join(result["ret"].keys()),
            )
        else:
            update_ret = await hub.exec.boto3.client.apigatewayv2.update_authorizer(
                ctx,
                ApiId=api_id,
                AuthorizerId=resource_id,
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            result["comment"] = (
                f"Updated parameters: " + ",".join(result["ret"].keys()),
            )

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.guardduty.create_detector
hub.exec.boto3.client.guardduty.untag_resource
hub.exec.boto3.client.guardduty.tag_resource
hub.exec.boto3.client.guardduty.get_detector
hub.exec.boto3.client.guardduty.list_detectors
hub.exec.boto3.client.guardduty.delete_detector
hub.exec.boto3.client.guardduty.update_detector
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    enable: bool = True,
    resource_id: str = None,
    client_token: str = None,
    finding_publishing_frequency: str = None,
    data_sources: make_dataclass(
        "DataSourceConfiguration",
        [
            (
                "S3Logs",
                make_dataclass("S3LogsConfiguration", [("Enable", bool)]),
                field(default=None),
            ),
            (
                "Kubernetes",
                make_dataclass(
                    "KubernetesConfiguration",
                    [
                        (
                            "AuditLogs",
                            make_dataclass(
                                "KubernetesAuditLogsConfiguration", [("Enable", bool)]
                            ),
                        )
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a single Amazon GuardDuty detector.
    A detector is a resource that represents the GuardDuty service.
    To start using GuardDuty, you must create a detector in each Region where you enable the service.
    You can have only one detector per account per Region.
    All data sources are enabled in a new detector by default.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): AWS DETECTOR id to identify the resource
        enable(bool, Default: True): A Boolean value that specifies whether the detector is to be enabled.
        client_token(Text, Optional): The idempotency token for the create request.
                                     This field is auto_populated if not provided.
        finding_publishing_frequency(Text, Optional): A value that specifies how frequently updated findings are exported.
        data_sources(Dict[str, Any], optional): Describes which data sources will be enabled for the detector. Defaults to None.
            * S3Logs (Dict[str, Any], optional): Describes whether S3 data event logs are enabled as a data source.
                * Enable (bool):  The status of S3 data event logs as a data source.
            * Kubernetes (Dict[str, Any], optional): Describes whether any Kubernetes logs are enabled as data sources.
                * AuditLogs (Dict[str, Any]): The status of Kubernetes audit logs as a data source.
                    * Enable (bool): The status of Kubernetes audit logs as a data source.
        tags(Dict[str, str], optional): The tags to be added to a new detector resource. Defaults to None.

    Request Syntax:
        [detector-resource-id]:
          aws.guaardduty.detector.present:
          - enable: True
          - client_token: 'string'
          - finding_publishing_frequency: 'string'
          - data_sources: dict
          - tags:
            - string: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            cebf7ced6562d943d61f76a915e32563:
                aws.guardduty.detector.present:
                    - enable: True
                    - finding_publishing_frequency: ONE_HOUR
                    - data_sources:
                        S3Logs:
                            Enable: true
                    - tags:
                        detector: my_detector

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.guardduty.get_detector(
            ctx, DetectorId=resource_id
        )

    if not before:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "enable": enable,
                    "finding_publishing_frequency": finding_publishing_frequency,
                    "data_sources": data_sources,
                    "tags": tags,
                },
            )
            result["comment"] = (f"Would create aws.guardduty.detector {name}",)
            return result
        try:
            ret = await hub.exec.boto3.client.guardduty.create_detector(
                ctx,
                Enable=enable,
                ClientToken=client_token,
                FindingPublishingFrequency=finding_publishing_frequency,
                DataSources=data_sources,
                Tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = result["comment"] + (f"Created '{name}'",)
            resource_id = ret["ret"]["DetectorId"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    else:
        try:
            account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
            account_id = account_details["ret"]["Account"]
            region_name = ctx["acct"]["region_name"]
            detector_arn = (
                "arn:aws:guardduty:"
                + region_name
                + ":"
                + account_id
                + ":detector/"
                + resource_id
            )
            result[
                "old_state"
            ] = await hub.tool.aws.guardduty.conversion_utils.convert_raw_detector_to_present(
                ctx, raw_resource=before, idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])
            update_ret = await hub.exec.aws.guardduty.detector.update_detector(
                ctx,
                before=before,
                detector_id=resource_id,
                finding_publishing_frequency=finding_publishing_frequency,
                data_sources=data_sources,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                if "finding_publishing_frequency" in update_ret["ret"]:
                    plan_state["finding_publishing_frequency"] = update_ret["ret"][
                        "finding_publishing_frequency"
                    ]
                if "data_sources" in update_ret["ret"]:
                    plan_state["data_sources"] = update_ret["ret"]["data_sources"]

            if (tags is not None) and (result["old_state"].get("tags", {}) != tags):
                update_tags_ret = (
                    await hub.exec.aws.guardduty.detector.update_detector_tags(
                        ctx=ctx,
                        resource_arn=detector_arn,
                        old_tags=result["old_state"].get("tags", {}),
                        new_tags=tags,
                    )
                )
                if not result["result"]:
                    result["comment"] = update_tags_ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = result["comment"] + update_tags_ret["comment"]
                result["result"] = result["result"] and update_tags_ret["result"]
                resource_updated = resource_updated or bool(update_tags_ret["ret"])
                if ctx.get("test", False) and update_tags_ret["ret"] is not None:
                    plan_state["tags"] = update_tags_ret["ret"].get("tags")
            if not resource_updated:
                result["comment"] = result["comment"] + (f"{name} already exists",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.guardduty.get_detector(
                ctx, DetectorId=resource_id
            )
            result[
                "new_state"
            ] = await hub.tool.aws.guardduty.conversion_utils.convert_raw_detector_to_present(
                ctx, raw_resource=after, idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes an Amazon GuardDuty detector that is specified by the detector ID.
    Args:
        resource_id(Text): The AWS ID of the detector.
        name(Text): A Idem name of the detector.

    Examples:

        .. code-block:: sls

            cebf7ced6562d943d61f76a915e32563:
              aws.guardduty.detector.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.guardduty.get_detector(
        ctx, DetectorId=resource_id
    )

    if not before:
        result["comment"] = (f"'{name}' already absent",)
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.guardduty.conversion_utils.convert_raw_detector_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        result["comment"] = (f"Would delete aws.guardduty.detector",)
        return result
    else:
        result[
            "old_state"
        ] = await hub.tool.aws.guardduty.conversion_utils.convert_raw_detector_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        try:
            ret = await hub.exec.boto3.client.guardduty.delete_detector(
                ctx, DetectorId=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Lists detectorIds of all the existing Amazon GuardDuty detector resources.
    Retrieves an Amazon GuardDuty detector specified by the detectorId.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.guardduty.detector
    """

    result = {}
    ret = await hub.exec.boto3.client.guardduty.list_detectors(ctx)

    if not ret["ret"]["DetectorIds"]:
        hub.log.debug(f"Could not list detector {ret['comment']}")
        return {}

    for resource_id in ret["ret"]["DetectorIds"]:
        detector_id = resource_id
        detector = await hub.exec.boto3.client.guardduty.get_detector(
            ctx, DetectorId=resource_id
        )

        resource_translated = await hub.tool.aws.guardduty.conversion_utils.convert_raw_detector_to_present(
            ctx, raw_resource=detector, idem_resource_name=detector_id
        )

        result[detector_id] = {
            "aws.guardduty.detector.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def update_budget(
    hub,
    ctx,
    account_id: str,
    before: Dict[str, Any],
    budget_limit: Dict[str, Any],
    planned_budget_limits: Dict[str, Any],
    cost_filters: Dict[str, Any],
    cost_types: Dict[str, Any],
    time_unit: str,
    time_period: Dict[str, Any],
    budget_type: str,
    auto_adjust_data: Dict[str, Any],
):

    """
    Updates a budget. You can change every part of a budget except for the budgetName
    and the calculatedSpend. When you modify a budget, the calculatedSpend drops to zero
    until Amazon Web Services has new usage data to use for forecasting.
    Args:
        hub: hub
        ctx: ctx
        account_id(str): AWS account id
        before: existing budget resource
        budget_limit(Dict[str, Any]): The total amount of cost that you want to track with your budget.
        planned_budget_limits(Dict[str, Any]): A map containing multiple BudgetLimit , including current or future limits.
        cost_filters(Dict[str, Any]): The cost filters, such as Region , Service , member account , Tag , or Cost Category , that are applied to a budget.
        cost_types(Dict[str, Any]): The types of costs that are included in this COST budget.
        time_unit(str): The length of time until a budget resets the actual and forecasted spend.
        time_period(Dict[str, Any]): The period of time that's covered by a budget. You set the start date and end date.
        budget_type(str): Specifies whether this budget tracks costs, usage, RI utilization, RI coverage, Savings Plans utilization, or Savings Plans coverage.
        auto_adjust_data(Dict[str, Any]): The parameters that determine the budget amount for an auto-adjusting budget.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None}
    """

    result = dict(comment=(), result=True, ret=None)

    # certain params are required in payload even if those are not changed.
    update_payload = {}
    payload = {}

    if budget_limit:
        payload["BudgetLimit"] = budget_limit
        if budget_limit != before.get("BudgetLimit"):
            update_payload["BudgetLimit"] = budget_limit

    if planned_budget_limits:
        payload["PlannedBudgetLimits"] = planned_budget_limits
        if planned_budget_limits != before.get("PlannedBudgetLimits"):
            update_payload["PlannedBudgetLimits"] = planned_budget_limits

    if cost_types and cost_types != before.get("CostTypes"):
        update_payload["CostTypes"] = cost_types
        payload["CostTypes"] = cost_types

    if time_period and (
        time_period["Start"] != str(before.get("TimePeriod")["Start"])
        or time_period["End"] != str(before.get("TimePeriod")["End"])
    ):
        update_payload["TimePeriod"] = time_period
        payload["TimePeriod"] = time_period

    if auto_adjust_data:
        payload["AutoAdjustData"] = auto_adjust_data
        if auto_adjust_data != before.get("AutoAdjustData"):
            update_payload["AutoAdjustData"] = auto_adjust_data

    # below params are mandatory, required in payload even if not changed.
    if cost_filters and cost_filters != before.get("CostFilters"):
        update_payload["CostFilters"] = cost_filters
    payload["CostFilters"] = cost_filters

    if time_unit and time_unit != before.get("TimeUnit"):
        update_payload["TimeUnit"] = time_unit
    payload["TimeUnit"] = time_unit

    if budget_type and budget_type != before.get("BudgetType"):
        update_payload["BudgetType"] = budget_type
    payload["BudgetType"] = budget_type

    payload["BudgetName"] = before.get("BudgetName")

    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.budgets.update_budget(
                ctx,
                AccountId=account_id,
                NewBudget=payload,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        result = update_result(result, update_payload)

    return result


def update_result(
    result: Dict[str, Any], update_payload: Dict[str, Any]
) -> Dict[str, Any]:
    parameters = OrderedDict(
        {
            "BudgetLimit": "budget_limit",
            "PlannedBudgetLimits": "planned_budget_limits",
            "CostFilters": "cost_filters",
            "CostTypes": "cost_types",
            "TimeUnit": "time_unit",
            "TimePeriod": "time_period",
            "BudgetType": "budget_type",
            "AutoAdjustData": "auto_adjust_data",
        }
    )

    for raw_parameter, present_parameter in parameters.items():
        if raw_parameter in update_payload:
            result["ret"][present_parameter] = update_payload[raw_parameter]
            result["comment"] += (
                f"Update {present_parameter}: {update_payload[raw_parameter]}",
            )
    return result
